"""Build script."""

import shutil
import sys
import tempfile
from pathlib import Path
from typing import Any, Dict

import skbuild

import grace.utils

PREFIX_DIR = Path(".") / "grace"
SRC_DIR = PREFIX_DIR / "extern"
PATCH_DIR = PREFIX_DIR / "patch"

# GRACE v2.2.1
GRACE_URL = "https://minami-home.kek.jp/grace221/grace.221.2006.0116.tgz"
GRACE_SHA256 = "a4971eddf65126a796c788661cdfa8c7cf738d91200d09a36999cc92baa943cf"
GRACE_SRC_DIR = SRC_DIR / "grace"
GRACE_PATCH_DIR = PATCH_DIR / "grace"

# grcext.py (2021-10-30)
GRCEXT_URL = (
    "https://gist.github.com/tueda/33861cb5469c7c431b1b99b943a725e7/archive/"
    "71a1dd22af90a0208bc19b94e61d55ccb29af57a.tar.gz"
)
GRCEXT_SHA256 = "aee80b53ed8bcc2bac2516849424078a6196130c03b87762f692fda396cc4a43"
GRCEXT_SRC_DIR = SRC_DIR / "grcext"
GRCEXT_PATCH_DIR = PATCH_DIR / "grcext"


def build(setup_kwargs: Dict[str, Any]) -> None:
    """Build the extensions."""
    # Check the main CMake file.
    cmake_lists = SRC_DIR / "CMakeLists.txt"
    if not cmake_lists.is_file():
        raise RuntimeError(f"{cmake_lists.absolute()} not found")

    # Prepare the source tree.
    if any(not p.exists() for p in (GRACE_SRC_DIR, GRCEXT_SRC_DIR)):
        reset()

    # Back up the default parameters.
    old_sys_argv = sys.argv
    old_skbuild_plat_name = skbuild.constants._SKBUILD_PLAT_NAME

    try:
        # Build (Release).
        sys.argv = old_sys_argv.copy()
        sys.argv.append("--build-type")
        sys.argv.append("Release")
        skbuild.constants._SKBUILD_PLAT_NAME = "Release" / Path(old_skbuild_plat_name)

        skbuild.setup(
            **setup_kwargs,
            script_args=["build_ext"],
            cmake_languages=("C", "Fortran"),
            cmake_source_dir=str(SRC_DIR),
        )

        release_install_dir = Path(skbuild.constants.CMAKE_INSTALL_DIR())

        # Build (Debug).
        sys.argv = old_sys_argv.copy()
        sys.argv.append("--build-type")
        sys.argv.append("Debug")
        skbuild.constants._SKBUILD_PLAT_NAME = "Debug" / Path(old_skbuild_plat_name)

        skbuild.setup(
            **setup_kwargs,
            script_args=["build_ext"],
            cmake_languages=("C", "Fortran"),
            cmake_source_dir=str(SRC_DIR),
        )

        debug_install_dir = Path(skbuild.constants.CMAKE_INSTALL_DIR())

        # Destination.
        grace_dest_dir = Path("grace")

        # Delete the build artifacts copied in previous runs, just in case.
        grace.utils.remove_files(grace_dest_dir, "bin")
        grace.utils.remove_files(grace_dest_dir, "lib")

        # Copy the build artifacts.
        grace.utils.copy_files(release_install_dir, grace_dest_dir, "bin")
        grace.utils.copy_files(release_install_dir, grace_dest_dir, "lib")

        # For the debug version.
        dest_dir = grace_dest_dir / "bin" / "debug"
        dest_dir.mkdir(parents=True, exist_ok=True)
        for src in (debug_install_dir / "bin").glob("*"):
            shutil.copy2(src, dest_dir / src.name)

        dest_dir = grace_dest_dir / "lib" / "debug"
        dest_dir.mkdir(parents=True, exist_ok=True)
        for src in (debug_install_dir / "lib").glob("*"):
            if src.is_file():
                shutil.copy2(src, dest_dir / src.name)

    finally:
        sys.argv = old_sys_argv
        skbuild.constants._SKBUILD_PLAT_NAME = old_skbuild_plat_name


def reset() -> None:
    """Reset the source tree."""
    grace.utils.download_archive(
        GRACE_URL,
        GRACE_SHA256,
        GRACE_SRC_DIR,
        clear=True,
        normalize_newlines=True,
        patch_dir=GRACE_PATCH_DIR,
    )

    grace.utils.download_archive(
        GRCEXT_URL,
        GRCEXT_SHA256,
        GRCEXT_SRC_DIR,
        clear=True,
        patch_dir=GRCEXT_PATCH_DIR,
    )


def diff() -> None:
    """Update the patch set."""
    with tempfile.TemporaryDirectory() as tmp_dir:
        temp_path = Path(tmp_dir)
        grace.utils.download_archive(
            GRACE_URL, GRACE_SHA256, temp_path, normalize_newlines=True
        )
        grace.utils.make_patch(
            temp_path, GRACE_SRC_DIR, GRACE_PATCH_DIR, normalize_newlines=True
        )

    with tempfile.TemporaryDirectory() as tmp_dir:
        temp_path = Path(tmp_dir)
        grace.utils.download_archive(GRCEXT_URL, GRCEXT_SHA256, temp_path)
        grace.utils.make_patch(temp_path, GRCEXT_SRC_DIR, GRCEXT_PATCH_DIR)


if __name__ == "__main__":
    build({})



"""Copying template files."""

import shutil
from pathlib import Path
from typing import Optional, Sequence

from .. import GRACE_ROOT


def copy_template(name: str, destdir: Optional[Path] = None) -> None:
    """Copy a template in.prc."""
    src = GRACE_ROOT / "lib" / "templates" / name / "in.prc"
    if destdir is None:
        destdir = Path.cwd()
    dest = destdir / "in.prc"
    shutil.copy(src, dest)


def list_templates() -> Sequence[str]:
    """List names of templates."""
    template_dir = GRACE_ROOT / "lib" / "templates"
    names = [str(f.relative_to(template_dir)) for f in template_dir.glob("*/*")]
    return tuple(names)



"""Version."""

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata  # type: ignore[no-redef]

try:
    __version__ = importlib_metadata.version("python-grace")
except importlib_metadata.PackageNotFoundError:
    # Undefined during installation (build.py).
    __version__ = None  # type: ignore[assignment]



"""Utility functions."""

import difflib
import filecmp
import hashlib
import os
import shutil
import urllib.request
import uuid
from pathlib import Path
from typing import Optional, Sequence

import appdirs
import patch

DEFAULT_IGNORE_PATTERNS = ("**/__pycache__/*", "**/*.log", "**/*.o", "**/*.swp")


try:
    # NOTE: when loaded from build.py, click may be not available.

    import click

    class GraceException(click.ClickException):
        """Exceptions that occur in Grace."""

        # NOTE: exceptions by ``raise GraceException(str)`` are automatically handled
        # in the Click framework.
        pass


except ModuleNotFoundError:

    class GraceException(Exception):  # type: ignore[no-redef]
        """Exceptions that occur in Grace."""

        pass


_cache_path: Optional[Path] = None


def get_cache_path() -> Path:
    """Return the cache path."""
    global _cache_path

    if _cache_path is None:
        # If $PYTHON_GRACE_CACHE_DIR is set, then use it.
        # Otherwise, use an OS-dependent cache directory.
        cache_dir = os.getenv("PYTHON_GRACE_CACHE_DIR")
        if not cache_dir:
            cache_dir = str(appdirs.user_cache_dir("python-grace"))
        _cache_path = Path(cache_dir)

    return _cache_path


def download_file(url: str, sha256: str) -> Path:
    """Download a file and return the path of the downloaded file."""
    # Check if the cache exists.

    url_hash = hashlib.sha256(url.encode()).hexdigest()

    cache_file = (
        get_cache_path() / "downloads" / url_hash[:2] / url_hash / Path(url).name
    )

    if not cache_file.exists():
        # Download the file contents.

        url_lower = url.lower()
        if not (url_lower.startswith("http:") or url_lower.startswith("https:")):
            raise ValueError("forbidden URL scheme: {url}")

        with urllib.request.urlopen(url) as f:  # noqa: S310  # only http/https
            data = f.read()

        # Use the "write-new-then-rename" idiom.

        temp_file = cache_file.parent / (cache_file.name + f".tmp{uuid.uuid4()}")

        temp_file.parent.mkdir(parents=True, exist_ok=True)

        try:
            with temp_file.open("wb") as f:
                f.write(data)

            os.replace(temp_file, cache_file)
        finally:
            if temp_file.exists():
                temp_file.unlink()

    # Check the sha256 hash of the downloaded file.

    file_hash = hashlib.sha256()

    with cache_file.open("rb") as f:
        while True:
            chunk = f.read(2048 * file_hash.block_size)
            if len(chunk) == 0:
                break
            file_hash.update(chunk)

    actual_sha256 = file_hash.hexdigest()
    if actual_sha256 != sha256:
        raise RuntimeError(
            f"wrong sha256 for file {cache_file} downloaded from {url},"
            f" expected: {sha256}, actual: {actual_sha256}"
        )

    return cache_file


def download_archive(
    url: str,
    sha256: str,
    extract_dir: Path,
    *,
    clear: bool = False,
    normalize_newlines: bool = False,
    patch_dir: Optional[Path] = None,
) -> Path:
    """Download an archive file and extract it to the given directory."""
    if clear and extract_dir.exists():
        shutil.rmtree(extract_dir)

    url_name = Path(url).name.lower()

    archive_suffixes = [
        ".tar.bz2",
        ".tar.gz",
        ".tar.xz",
        ".tar",
        ".tbz",
        ".tgz",
        ".zip",
    ]

    if any(url_name.endswith(suffix) for suffix in archive_suffixes):
        cache_file = download_file(url, sha256)

        temp_dir = extract_dir.parent / (extract_dir.name + f".tmp{uuid.uuid4()}")

        temp_dir.mkdir(parents=True, exist_ok=False)
        try:
            shutil.unpack_archive(
                os.fspath(cache_file), extract_dir=os.fspath(temp_dir)
            )  # fspath for python<3.7

            temp_root_dir = temp_dir
            while True:
                items = list(temp_root_dir.iterdir())
                if len(items) == 1 and items[0].is_dir():
                    temp_root_dir = items[0]
                else:
                    break

            move_tree(temp_root_dir, extract_dir)
        finally:
            if temp_dir.exists():
                shutil.rmtree(temp_dir)
    else:
        raise ValueError(f"{url} does not look like an archive file")

    if normalize_newlines:
        normalize_newlines_intree(extract_dir)

    if patch_dir:
        apply_patch(extract_dir, patch_dir)

    return extract_dir


def move_tree(src_dir: Path, dest_dir: Path) -> Path:
    """Move an entire directory tree to another."""
    if not src_dir.is_dir():
        raise ValueError(f"directory {src_dir} does not exist")
    if dest_dir.exists():
        shutil.rmtree(dest_dir)
    shutil.move(str(src_dir), str(dest_dir))  # str(...) for Python < 3.9
    return dest_dir


def normalize_newlines_intext(text: str) -> str:
    """Normalize new lines in the given text."""
    # This corresponds to:
    #   end_of_line = lf
    #   insert_final_newline = true
    #   trim_trailing_whitespace = true
    text = "\n".join(line.rstrip() for line in text.splitlines())
    if text and text[-1] != "\n":
        text += "\n"
    return text


def normalize_newlines_intree(target_dir: Path) -> None:
    """Normalize new lines in text files in an entire directory tree."""
    for path in target_dir.iterdir():
        if path.is_dir():
            normalize_newlines_intree(path)
            continue

        try:
            # We expect reading binary files or non-UTF8 files fails.
            text1 = path.read_text()
            text2 = normalize_newlines_intext(text1)

            if text1 != text2:
                path.write_text(text2)
        except UnicodeDecodeError:
            pass


def make_patch(
    original_dir: Path,
    changed_dir: Path,
    patch_dir: Path,
    *,
    ignore_patterns: Optional[Sequence[str]] = None,
    normalize_newlines: bool = False,
) -> None:
    """Generate a patch set."""
    if normalize_newlines:
        normalize = normalize_newlines_intext
    else:

        def normalize(text: str) -> str:
            return text

    def make_patch_file(relative_path: Path) -> None:
        file1 = original_dir / relative_path
        file2 = changed_dir / relative_path

        def do_patch_file() -> None:
            if filecmp.cmp(file1, file2):
                return

            patch_file = patch_dir / relative_path
            patch_file = patch_file.parent / f"{patch_file.name}.patch"

            lines1 = normalize(file1.read_text()).splitlines(keepends=True)
            lines2 = normalize(file2.read_text()).splitlines(keepends=True)

            diffs = list(
                difflib.unified_diff(
                    lines1,
                    lines2,
                    str(Path("a") / relative_path),
                    str(Path("b") / relative_path),
                )
            )

            if not diffs:
                return

            patch_file.parent.mkdir(parents=True, exist_ok=True)
            patch_file.write_text("".join(diffs))

        def do_new_file() -> None:
            patch_file = patch_dir / relative_path
            patch_file = patch_file.parent / patch_file.name

            patch_file.parent.mkdir(parents=True, exist_ok=True)
            patch_file.write_text(normalize(file2.read_text()))

        if file1.is_file():
            if file2.is_file():
                do_patch_file()
                return

        if not file1.exists():
            if file2.is_file():
                do_new_file()
                return

        raise RuntimeError(f"diff between {file1} and {file2} not supported")

    # Ensure that the patch directory is empty.

    if patch_dir.exists():
        shutil.rmtree(patch_dir)
    patch_dir.mkdir(parents=True)

    # Collect all files in the two directories.

    if ignore_patterns:
        ignore_patterns = DEFAULT_IGNORE_PATTERNS + tuple(ignore_patterns)
    else:
        ignore_patterns = DEFAULT_IGNORE_PATTERNS

    items = [
        path.relative_to(original_dir)
        for path in original_dir.glob("**/*")
        if path.is_file()
        and all(not path.match(pattern) for pattern in ignore_patterns)
    ] + [
        path.relative_to(changed_dir)
        for path in changed_dir.glob("**/*")
        if path.is_file()
        and all(not path.match(pattern) for pattern in ignore_patterns)
    ]

    # Generate a patch for each file.

    for relative_path in set(items):
        make_patch_file(relative_path)


def apply_patch(target_dir: Path, patch_dir: Path) -> None:
    """Apply a patch set."""
    # Copy all new files.
    ignore_patterns = DEFAULT_IGNORE_PATTERNS + ("**/*.patch",)
    new_files = [
        path
        for path in patch_dir.glob("**/*")
        if path.is_file()
        and all(not path.match(pattern) for pattern in ignore_patterns)
    ]
    for new_file in new_files:
        target_file = target_dir / new_file.relative_to(patch_dir)
        target_file = target_file.parent / target_file.name
        target_file.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy2(new_file, target_file)

    # Apply all patch files.
    patch_files = patch_dir.glob("**/*.patch")
    for patch_file in patch_files:
        patch_set = patch.fromfile(patch_file)
        if not patch_set:
            raise RuntimeError(f"failed to load {patch_file}")
        if not patch_set.apply(strip=0, root=target_dir):
            raise RuntimeError(f"failed apply {patch_file}")


def remove_files(target_dir: Path, pattern: str) -> None:
    """Delete files matched with a glob pattern in a directory tree."""
    for path in target_dir.glob(pattern):
        if path.is_dir():
            shutil.rmtree(path)
        else:
            path.unlink()


def copy_files(src_dir: Path, dest_dir: Path, pattern: str) -> None:
    """Copy files matched with a glob pattern in a directory tree to another."""
    for src in src_dir.glob(pattern):
        dest = dest_dir / src.relative_to(src_dir)
        if src.is_dir():
            copy_files(src, dest, "*")
        else:
            dest.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy2(src, dest)



"""GRACE commands."""

from . import template as template  # noqa: F401  # reexport for mypy
from .raw import RawCommand
from .raw import raw_commands as raw_commands  # noqa: F401  # reexport for mypy

gracefig = RawCommand("gracefig")
grc = RawCommand("grc")
grccut = RawCommand("grccut")
grcdraw = RawCommand("grcdraw")
grcext = RawCommand("grcext", python=True)
grcfort = RawCommand("grcfort")
grcmdl = RawCommand("grcmdl")
grcmom = RawCommand("grcmom")
grcplot = RawCommand("grcplot")
grcprc = RawCommand("grcprc")
qcdcut = RawCommand("qcdcut")
tread = RawCommand("tread")



"""Main program."""

from typing import List, Sequence

import click

from . import __version__, commands
from .click_ext import OrderedGroup


@click.group(
    cls=OrderedGroup,
    invoke_without_command=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)
@click.pass_context
@click.option("--version", is_flag=True, help="Show the version.")
@click.option("--debug/--no-debug", is_flag=True, help="Enable the debugging mode.")
def main(  # noqa: D103  # to suppress the help message
    ctx: click.Context, version: bool, debug: bool
) -> None:
    if ctx.invoked_subcommand is None:
        if version:
            click.echo(f"python-grace {__version__}")
        else:
            click.echo(main.get_help(ctx))
    else:
        ctx.ensure_object(dict)
        ctx.obj["DEBUG"] = debug


def _complete_template_name(
    ctx: click.Context, param: str, incomplete: str
) -> List[str]:
    names = commands.template.list_templates()
    return [k for k in names if k.startswith(incomplete)]


@main.command(no_args_is_help=True)
@click.argument("name", shell_complete=_complete_template_name)
def template(name: str) -> None:
    """Copy a template in.prc to the current directory."""
    commands.template.copy_template(name)


def _add_template_help() -> None:
    if not template.help:
        return

    names = commands.template.list_templates()

    if names:
        s = ", ".join(names)
        template.help += f"\n\nNAME must be one of: {s}."


_add_template_help()


def _add_raw_commands() -> None:
    args = click.Argument(["args"], required=False, nargs=-1)
    context_settings = {"ignore_unknown_options": True, "help_option_names": []}

    def make_callback(c: click.decorators.F) -> click.decorators.F:
        @click.pass_context
        def callback(ctx: click.Context, args: Sequence[str]) -> None:
            c(args, ctx.obj["DEBUG"])

        return callback  # type: ignore[return-value]

    for c in commands.raw_commands.values():
        if not c.available:
            continue
        name = c.name

        cmd = click.Command(
            name,
            context_settings=context_settings,
            callback=make_callback(c),
            params=[args],
            help=f"Run {name}.",
        )
        main.add_command(cmd)


_add_raw_commands()



"""Raw commands."""

import os
import re
import subprocess
import sys
from collections import OrderedDict
from pathlib import Path
from typing import Sequence

from .. import GRACE_ROOT
from ..utils import GraceException

raw_commands: "OrderedDict[str, RawCommand]" = OrderedDict()


class RawCommand:
    """Raw command object."""

    def __init__(self, cmd: str, python: bool = False) -> None:
        """Construct a raw command object."""
        self._cmd = cmd
        self._python = python
        raw_commands[cmd] = self

    def __call__(self, args: Sequence[str], debug: bool) -> None:
        """Invoke the raw command."""
        if debug:
            cmd_args = [str(GRACE_ROOT / "bin" / "debug" / self._cmd)]
        else:
            cmd_args = [str(GRACE_ROOT / "bin" / self._cmd)]

        if self._python:
            cmd_args = [sys.executable] + cmd_args

        cmd_args += list(args)

        # We need to give the environment variables GRCMODEL and KINEMPATH.

        model_path = GRACE_ROOT / "lib" / "model"
        kinem_path = GRACE_ROOT / "lib" / "dbkinem"

        env = os.environ.copy()
        env["GRCMODEL"] = f".:{model_path}"
        env["KINEMPATH"] = f".:{kinem_path}"

        try:
            subprocess.run(cmd_args, check=True, env=env)  # noqa: S603
        except subprocess.CalledProcessError as e:
            raise GraceException(
                f"command {e.cmd} returned non-zero exit status {e.returncode}"
            )

        if self._cmd == "grcfort":
            patch_makefile()

    @property
    def name(self) -> str:
        """Return the command name."""
        return self._cmd

    @property
    def available(self) -> bool:
        """Return `True` if the command is available."""
        # This assumes that bin and bin/debug have the same executables.
        return is_raw_command_available(self._cmd)


def patch_makefile() -> None:
    """Patch Makefile."""
    makefile = Path("Makefile")

    if not makefile.exists():
        return

    lines = makefile.read_text().splitlines()

    # We need to provide GRACEROOT.

    if any(re.match(r"^\s*GRACEROOT\s*=", line) for line in lines):
        return

    for i, line in enumerate(lines):
        # Insert a line just before GRACEDIR.
        if re.match(r"^\s*GRACEDIR\s*=", line):
            lines.insert(i, f"GRACEROOT = {str(GRACE_ROOT)}")
            break
    else:
        raise RuntimeError("failed to patch Makefile")

    makefile.write_text("\n".join(lines) + "\n")


def is_raw_command_available(cmd: str, debug: bool = False) -> bool:
    """Return `True` if the given raw command is actually available."""
    if debug:
        return (GRACE_ROOT / "bin" / "debug" / cmd).is_file()
    else:
        return (GRACE_ROOT / "bin" / cmd).is_file()



"""Click extensions."""

from collections import OrderedDict
from typing import Any, Dict, List, Optional

import click


class OrderedGroup(click.Group):
    """Group preserving the order of subcommands."""

    # See:
    # https://github.com/pallets/click/issues/513#issuecomment-504158316
    # https://stackoverflow.com/a/58323807

    def __init__(
        self,
        name: Optional[str] = None,
        commands: Optional[Dict[str, click.Command]] = None,
        **attrs: Dict[str, Any],
    ) -> None:
        """Construct a group."""
        if commands is None:
            commands = OrderedDict()
        elif not isinstance(commands, OrderedDict):
            commands = OrderedDict(commands)
        super(OrderedGroup, self).__init__(name=name, commands=commands, **attrs)

    def list_commands(self, ctx: click.Context) -> List[str]:
        """Return a list of subcommands in the order they should appear."""
        return list(self.commands.keys())
